<?php

declare(strict_types=1);

namespace Drupal\user_registration_reminder;

use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\user\UserDataInterface;
use Drupal\user\UserInterface;
use Symfony\Component\Routing\Exception\InvalidParameterException;

/**
 * Decorates Drupal User entities for User Registration Reminder.
 *
 * These users are typically user accounts that were never activated.
 */
class UserDecorator {

  /**
   * The recent escalation step for this user.
   */
  const KEY_ESCALATION_STEP = 'escalation.step';

  /**
   * The recent escalation timestamp for this user.
   */
  const KEY_ESCALATION_TIMESTAMP = 'escalation.timestamp';

  use StringTranslationTrait;

  /**
   * Constructs a Manager object.
   */
  public function __construct(
    protected readonly UserInterface $user,
    protected readonly UserDataInterface $userData,
  ) {
    // Validate user:
    if (!empty($user->getLastLoginTime())) {
      throw new \Exception("Users who logged in should not be handled here. Given login time is '{$user->getLastLoginTime()}'. Something is wrong...");
    }
    if (!$user->isActive()) {
      throw new \Exception("Inactive users should not be handled here. Something is wrong...");
    }
  }

  /**
   * Helper method to return the module specific user data.
   *
   * For yet unprocessed users this is always empty.
   *
   * @param string|null $key
   *   Specific key, otherwise the whole array is returned.
   *
   * @return mixed
   *   The resulting user data.
   */
  protected function getUserData(?string $key = NULL): mixed {
    return $this->userData->get('user_registration_reminder', $this->user->id(), $key) ?? [];
  }

  /**
   * Returns the most recent escalation step count from user data.
   *
   * Returns 0 if the user has not been processed yet.
   *
   * @return int
   *   The most recent escalation step count.
   */
  public function getRecentEscalationStep(): int {
    return (int) $this->getUserData(self::KEY_ESCALATION_STEP) ?? 0;
  }

  /**
   * The timestamp of the most recent escalation. Zero if fresh.
   *
   * Returns 0 if the user has not been processed yet.
   *
   * @return int
   *   The timestamp of the most recent escalation run.
   */
  public function getRecentEscalationTimestamp(): int {
    return (int) $this->getUserData(self::KEY_ESCALATION_TIMESTAMP) ?? 0;
  }

  /**
   * Increases the escalation step and set the date in user data.
   *
   * Sets the internal user data for this module in the user data, so that
   * we can check in which escalation step this user currently is and
   * how to proceed with it.
   */
  public function increaseEscalationStep(): void {
    $this->userData->set('user_registration_reminder', $this->user->id(), self::KEY_ESCALATION_STEP, $this->getRecentEscalationStep() + 1);
    $this->userData->set('user_registration_reminder', $this->user->id(), self::KEY_ESCALATION_TIMESTAMP, time());
  }

  /**
   * Executes the given "final action" on the user profile.
   *
   * This should be called after all escalation steps have been executed.
   *
   * @param string $action
   *   The final action to execute.
   *
   * @return bool
   *   Returns true if an action has
   */
  public function executeFinalAction(string $action): bool {
    switch ($action) {
      case Manager::FINAL_ACTION_BLOCK:
        $this->getUser()->block()->save();
        return TRUE;

      case Manager::FINAL_ACTION_DELETE:
        user_cancel([], $this->user->id(), 'user_cancel_delete');
        return TRUE;

      case Manager::FINAL_ACTION_NONE:
        // Do nothing. Not that these users will be looped every time
        // because user.data can't be sorted out well as it's a
        // serialized array and so we have no SQL flag.
        // That's why the other options are preferred.
        return TRUE;

      default:
        throw new InvalidParameterException("Unknown action: {$action}.");
    }
  }

  /**
   * Returns the decorated user object.
   *
   * @return \Drupal\user\Entity\UserInterface
   *   The user object.
   */
  public function getUser(): UserInterface {
    return $this->user;
  }

}
